@@ -1,17 +1,13 @@
 #ifndef MUPDF_PDF_DOCUMENT_H
 #define MUPDF_PDF_DOCUMENT_H
 
-typedef struct pdf_lexbuf_s pdf_lexbuf;
-typedef struct pdf_lexbuf_large_s pdf_lexbuf_large;
-typedef struct pdf_xref_s pdf_xref;
-typedef struct pdf_ocg_descriptor_s pdf_ocg_descriptor;
-typedef struct pdf_portfolio_s pdf_portfolio;
+typedef struct pdf_xref pdf_xref;
+typedef struct pdf_ocg_descriptor pdf_ocg_descriptor;
 
-typedef struct pdf_page_s pdf_page;
-typedef struct pdf_annot_s pdf_annot;
-typedef struct pdf_annot_s pdf_widget;
-typedef struct pdf_hotspot_s pdf_hotspot;
-typedef struct pdf_js_s pdf_js;
+typedef struct pdf_page pdf_page;
+typedef struct pdf_annot pdf_annot;
+typedef struct pdf_annot pdf_widget;
+typedef struct pdf_js pdf_js;
 
 enum
 {
@@ -19,43 +15,37 @@ enum
 	PDF_LEXBUF_LARGE = 65536
 };
 
-struct pdf_lexbuf_s
+typedef struct
 {
-	int size;
-	int base_size;
-	int len;
+	size_t size;
+	size_t base_size;
+	size_t len;
 	int64_t i;
 	float f;
 	char *scratch;
 	char buffer[PDF_LEXBUF_SMALL];
-};
+} pdf_lexbuf;
 
-struct pdf_lexbuf_large_s
+typedef struct
 {
 	pdf_lexbuf base;
 	char buffer[PDF_LEXBUF_LARGE - PDF_LEXBUF_SMALL];
-};
-
-struct pdf_hotspot_s
-{
-	int num;
-	int state;
-};
+} pdf_lexbuf_large;
 
 /*
 	Document event structures are mostly opaque to the app. Only the type
 	is visible to the app.
 */
-typedef struct pdf_doc_event_s pdf_doc_event;
+typedef struct pdf_doc_event pdf_doc_event;
 
 /*
-	pdf_doc_event_cb: the type of function via which the app receives
+	the type of function via which the app receives
 	document events.
 */
 typedef void (pdf_doc_event_cb)(fz_context *ctx, pdf_document *doc, pdf_doc_event *event, void *data);
 
 /*
-	pdf_open_document: Open a PDF document.
+	Open a PDF document.
 
 	Open a PDF document by reading its cross reference table, so
 	MuPDF can locate PDF objects inside the file. Upon an broken
@@ -74,7 +64,7 @@ typedef void (pdf_doc_event_cb)(fz_context *ctx, pdf_document *doc, pdf_doc_even
 pdf_document *pdf_open_document(fz_context *ctx, const char *filename);
 
 /*
-	pdf_open_document_with_stream: Opens a PDF document.
+	Opens a PDF document.
 
 	Same as pdf_open_document, but takes a stream instead of a
 	filename to locate the PDF document to open. Increments the
@@ -85,39 +75,32 @@ pdf_document *pdf_open_document(fz_context *ctx, const char *filename);
 pdf_document *pdf_open_document_with_stream(fz_context *ctx, fz_stream *file);
 
 /*
-	pdf_drop_document: Closes and frees an opened PDF document.
+	Closes and frees an opened PDF document.
 
 	The resource store in the context associated with pdf_document
 	is emptied.
 */
 void pdf_drop_document(fz_context *ctx, pdf_document *doc);
 
-/*
-	pdf_keep_document: Keep a reference to an open document.
-*/
 pdf_document *pdf_keep_document(fz_context *ctx, pdf_document *doc);
 
 /*
-	pdf_specifics: down-cast a fz_document to a pdf_document.
+	down-cast a fz_document to a pdf_document.
 	Returns NULL if underlying document is not PDF
 */
 pdf_document *pdf_specifics(fz_context *ctx, fz_document *doc);
 
 /*
-	pdf_document_from_fz_document,
-	pdf_page_from_fz_page,
-	pdf_annot_from_fz_annot:
-		Down-cast generic fitz objects into pdf specific variants.
-		Returns NULL if the objects are not from a PDF document.
+	Down-cast generic fitz objects into pdf specific variants.
+	Returns NULL if the objects are not from a PDF document.
 */
 pdf_document *pdf_document_from_fz_document(fz_context *ctx, fz_document *ptr);
 pdf_page *pdf_page_from_fz_page(fz_context *ctx, fz_page *ptr);
-pdf_annot *pdf_annot_from_fz_annot(fz_context *ctx, fz_annot *ptr);
 
 int pdf_needs_password(fz_context *ctx, pdf_document *doc);
 
 /*
-	pdf_authenticate_password: Attempt to authenticate a
+	Attempt to authenticate a
 	password.
 
 	Returns 0 for failure, non-zero for success.
@@ -135,13 +118,15 @@ int pdf_lookup_metadata(fz_context *ctx, pdf_document *doc, const char *key, cha
 fz_outline *pdf_load_outline(fz_context *ctx, pdf_document *doc);
 
 /*
-	pdf_count_layer_configs: Get the number of layer
-	configurations defined in this document.
+	Get the number of layer configurations defined in this document.
 
 	doc: The document in question.
 */
 int pdf_count_layer_configs(fz_context *ctx, pdf_document *doc);
 
+void pdf_invalidate_xfa(fz_context *ctx, pdf_document *doc);
+
+
 typedef struct
 {
 	const char *name;
@@ -149,8 +134,7 @@ typedef struct
 } pdf_layer_config;
 
 /*
-	pdf_layer_config_info: Fetch the name (and
-	optionally creator) of the given layer config.
+	Fetch the name (and optionally creator) of the given layer config.
 
 	doc: The document in question.
 
@@ -164,7 +148,7 @@ typedef struct
 void pdf_layer_config_info(fz_context *ctx, pdf_document *doc, int config_num, pdf_layer_config *info);
 
 /*
-	pdf_select_layer_config: Set the current configuration.
+	Set the current configuration.
 	This updates the visibility of the optional content groups
 	within the document.
 
@@ -176,16 +160,15 @@ void pdf_layer_config_info(fz_context *ctx, pdf_document *doc, int config_num, p
 void pdf_select_layer_config(fz_context *ctx, pdf_document *doc, int config_num);
 
 /*
-	pdf_count_layer_config_ui: Returns the number of entries in the
-	'UI' for this layer configuration.
+	Returns the number of entries in the 'UI' for this layer configuration.
 
 	doc: The document in question.
 */
 int pdf_count_layer_config_ui(fz_context *ctx, pdf_document *doc);
 
 /*
-	pdf_select_layer_ui: Select a checkbox/radiobox
-	within the 'UI' for this layer configuration.
+	Select a checkbox/radiobox within the 'UI' for this layer
+	configuration.
 
 	Selecting a UI entry that is a radiobox may disable
 	other UI entries.
@@ -198,8 +181,7 @@ int pdf_count_layer_config_ui(fz_context *ctx, pdf_document *doc);
 void pdf_select_layer_config_ui(fz_context *ctx, pdf_document *doc, int ui);
 
 /*
-	pdf_deselect_layer_ui: Select a checkbox/radiobox
-	within the 'UI' for this layer configuration.
+	Select a checkbox/radiobox within the 'UI' for this layer configuration.
 
 	doc: The document in question.
 
@@ -209,8 +191,7 @@ void pdf_select_layer_config_ui(fz_context *ctx, pdf_document *doc, int ui);
 void pdf_deselect_layer_config_ui(fz_context *ctx, pdf_document *doc, int ui);
 
 /*
-	pdf_toggle_layer_config_ui: Toggle a checkbox/radiobox
-	within the 'UI' for this layer configuration.
+	Toggle a checkbox/radiobox within the 'UI' for this layer configuration.
 
 	Toggling a UI entry that is a radiobox may disable
 	other UI entries.
@@ -239,8 +220,7 @@ typedef struct
 } pdf_layer_config_ui;
 
 /*
-	pdf_layer_config_ui_info: Get the info for a given
-	entry in the layer config ui.
+	Get the info for a given entry in the layer config ui.
 
 	doc: The document in question.
 
@@ -253,347 +233,66 @@ typedef struct
 void pdf_layer_config_ui_info(fz_context *ctx, pdf_document *doc, int ui, pdf_layer_config_ui *info);
 
 /*
-	pdf_set_layer_config_as_default: Write the current layer
-	config back into the document as the default state.
+	Write the current layer config back into the document as the default state.
 */
 void pdf_set_layer_config_as_default(fz_context *ctx, pdf_document *doc);
 
-/*
-	PDF portfolios (or collections) are embedded files. They can
-	be thought of as tables of information, with an embedded
-	file per row. For instance a PDF portfolio of an email box might
-	contain:
-
-			From	To	Cc	Date
-	message1.pdf	...	...	...	...
-	message2.pdf	...	...	...	...
-
-	etc. The details of the 'column headings' are known as the Schema.
-	This includes the order to use for the headings.
-
-	Each row in the table is a portfolio (or collection) entry.
-*/
-
-/*
-	pdf_count_portfolio_schema: Get the number of entries in the
-	portfolio schema used in this document.
-
-	doc: The document in question.
-*/
-int pdf_count_portfolio_schema(fz_context *ctx, pdf_document *doc);
-
-typedef enum
-{
-	PDF_SCHEMA_NUMBER,
-	PDF_SCHEMA_SIZE,
-	PDF_SCHEMA_TEXT,
-	PDF_SCHEMA_DATE,
-	PDF_SCHEMA_DESC,
-	PDF_SCHEMA_MODDATE,
-	PDF_SCHEMA_CREATIONDATE,
-	PDF_SCHEMA_FILENAME,
-	PDF_SCHEMA_UNKNOWN
-} pdf_portfolio_schema_type;
-
-typedef struct
-{
-	pdf_portfolio_schema_type type;
-	int visible;
-	int editable;
-	pdf_obj *name;
-} pdf_portfolio_schema;
-
-/*
-	pdf_portfolio_schema_info: Fetch information about a given
-	portfolio schema entry.
-
-	doc: The document in question.
-
-	entry: A value in the 0..n-1 range, where n is the
-	value returned from pdf_count_portfolio_schema.
-
-	info: Pointer to structure to fill in. Pointers within
-	this structure may be set to NULL if no information is
-	available.
-*/
-void pdf_portfolio_schema_info(fz_context *ctx, pdf_document *doc, int entry, pdf_portfolio_schema *info);
-
-/*
-	pdf_reorder_portfolio_schema: Reorder the portfolio schema.
-
-	doc: The document in question.
-
-	entry: A value in the 0..n-1 range, where n is the
-	value returned from pdf_count_portfolio_schema - the
-	position of the entry to move.
-
-	new_pos: A value in the 0..n-1 range, where n is the
-	value returned from pdf_count_portfolio_schema - the
-	position to move the entry to.
-*/
-void pdf_reorder_portfolio_schema(fz_context *ctx, pdf_document *doc, int entry, int new_pos);
-
-/*
-	pdf_rename_portfolio_schema: rename a given portfolio
-	schema entry.
-
-	doc: The document in question.
-
-	entry: The entry to renumber.
-
-	name: The new name for the portfolio schema
-
-	name_len: The byte length of the name.
-*/
-void pdf_rename_portfolio_schema(fz_context *ctx, pdf_document *doc, int entry, const char *name, int name_len);
-
-/*
-	pdf_delete_portfolio_schema: delete a given portfolio
-	schema entry.
-
-	doc: The document in question.
-
-	entry: The entry to delete.
-*/
-void pdf_delete_portfolio_schema(fz_context *ctx, pdf_document *doc, int entry);
-
-/*
-	pdf_add_portfolio_schema: Add a new portfolio schema
-	entry.
-
-	doc: The document in question.
-
-	entry: The point in the ordering at which to insert the new
-	schema entry.
-
-	info: Details of the schema entry.
-*/
-void pdf_add_portfolio_schema(fz_context *ctx, pdf_document *doc, int entry, const pdf_portfolio_schema *info);
-
-/*
-	pdf_count_portfolio_entries: Get the number of portfolio entries
-	in this document.
-
-	doc: The document in question.
-*/
-int pdf_count_portfolio_entries(fz_context *ctx, pdf_document *doc);
-
-/*
-	pdf_portfolio_entry: Create a buffer containing
-	a decoded portfolio entry.
-
-	doc: The document in question.
-
-	entry: A value in the 0..m-1 range, where m is the
-	value returned from pdf_count_portfolio_entries.
-
-	Returns a buffer containing the decoded portfolio
-	entry. Ownership of the buffer passes to the caller.
-*/
-fz_buffer *pdf_portfolio_entry(fz_context *ctx, pdf_document *doc, int entry);
-
-/*
-	pdf_portfolio_entry_obj_name: Retrieve the object and
-	name of a given portfolio entry.
-
-	doc: The document in question.
-
-	entry: A value in the 0..m-1 range, where m is the
-	value returned from pdf_count_portfolio_entries.
-
-	name: Pointer to a place to store the pointer to the
-	object representing the name. This is a borrowed
-	reference - do not drop it.
-
-	Returns a pointer to the pdf_object representing the
-	object. This is a borrowed reference - do not drop
-	it.
-*/
-pdf_obj *pdf_portfolio_entry_obj_name(fz_context *ctx, pdf_document *doc, int entry, pdf_obj **name);
-
-/*
-	pdf_portfolio_entry_obj: Retrieve the object
-	representing a given portfolio entry.
-
-	doc: The document in question.
-
-	entry: A value in the 0..m-1 range, where m is the
-	value returned from pdf_count_portfolio_entries.
-
-	Returns a pointer to the pdf_object representing the
-	object. This is a borrowed reference - do not drop
-	it.
-*/
-pdf_obj *pdf_portfolio_entry_obj(fz_context *ctx, pdf_document *doc, int entry);
-
-/*
-	pdf_portfolio_entry_name: Retrieve the name of
-	a given portfolio entry.
-
-	doc: The document in question.
-
-	entry: A value in the 0..m-1 range, where m is the
-	value returned from pdf_count_portfolio_entries.
-
-	name: Pointer to a place to store the pointer to the
-	object representing the name. This is a borrowed
-	reference - do not drop it.
-
-	Returns a pointer to the pdf_object representing the
-	name of the entry. This is a borrowed reference - do not drop
-	it.
-*/
-pdf_obj *pdf_portfolio_entry_name(fz_context *ctx, pdf_document *doc, int entry);
-
-/*
-	pdf_portfolio_entry_info: Fetch information about a given
-	portfolio entry.
-
-	doc: The document in question.
-
-	entry: A value in the 0..m-1 range, where m is the
-	value returned from pdf_count_portfolio_entries.
-
-	info: Pointer to structure to fill in. Pointers within
-	this structure may be set to NULL if no information is
-	available.
-*/
-pdf_obj *pdf_portfolio_entry_info(fz_context *ctx, pdf_document *doc, int entry, int schema_entry);
-
-/*
-	pdf_add_portfolio_entry: Add a new portfolio entry.
-
-	doc: The document in question.
-
-	name: The name to use for this entry (as used in the
-	PDF name tree for the collection).
-
-	name_len: The byte length of name.
-
-	desc: The description to use for this entry (as used
-	in the 'Desc' entry in the Collection entry).
-
-	desc_len: The byte length of desc.
-
-	filename: The filename to use for this entry (as used
-	in the 'F' entry in the collection entry).
-
-	filename_len: The byte length of filename.
-
-	unifilename: The filename to use for this entry (as used
-	in the 'UF' entry in the collection entry).
-
-	unifilename_len: The byte length of unifilename.
-
-	buf: The buffer containing the embedded file to add.
-
-	Returns the entry number for this new entry.
-*/
-int pdf_add_portfolio_entry(fz_context *ctx, pdf_document *doc,
-				const char *name, int name_len,
-				const char *desc, int desc_len,
-				const char *filename, int filename_len,
-				const char *unifile, int unifile_len, fz_buffer *buf);
-
-/*
-	pdf_set_portfolio_entry_info: Set part of the entry
-	information for a given portfolio entry.
-
-	doc: The document in question.
-
-	entry: The portfolio entry to set information for.
-	In the range 0..m-1, where m is the value returned
-	from pdf_count_portfolio_entries.
-
-	schema_entry: Which schema entry to set (in the
-	range 0..n-1, where n is the value returned from
-	pdf_count_portfolio_schema.
-
-	data: The value to set.
-*/
-void pdf_set_portfolio_entry_info(fz_context *ctx, pdf_document *doc, int entry, int schema_entry, pdf_obj *data);
-
 /*
 	Determine whether changes have been made since the
 	document was opened or last saved.
 */
 int pdf_has_unsaved_changes(fz_context *ctx, pdf_document *doc);
 
-enum pdf_signature_error
-{
-	PDF_SIGNATURE_ERROR_OKAY,
-	PDF_SIGNATURE_ERROR_NO_SIGNATURES,
-	PDF_SIGNATURE_ERROR_NO_CERTIFICATE,
-	PDF_SIGNATURE_ERROR_DOCUMENT_CHANGED,
-	PDF_SIGNATURE_ERROR_SELF_SIGNED,
-	PDF_SIGNATURE_ERROR_SELF_SIGNED_IN_CHAIN,
-	PDF_SIGNATURE_ERROR_NOT_TRUSTED,
-	PDF_SIGNATURE_ERROR_UNKNOWN
-};
-
-typedef struct pdf_pkcs7_designated_name_s
-{
-	char *cn;
-	char *o;
-	char *ou;
-	char *email;
-	char *c;
-}
-pdf_pkcs7_designated_name;
+/*
+	Determine if this PDF has been repaired since opening.
+*/
+int pdf_was_repaired(fz_context *ctx, pdf_document *doc);
 
 /* Object that can perform the cryptographic operation necessary for document signing */
-typedef struct pdf_pkcs7_signer_s pdf_pkcs7_signer;
-
-/* Increment the reference count for a signer object */
-typedef pdf_pkcs7_signer *(pdf_pkcs7_keep_fn)(pdf_pkcs7_signer *signer);
-
-/* Drop a reference for a signer object */
-typedef void (pdf_pkcs7_drop_fn)(pdf_pkcs7_signer *signer);
-
-/* Obtain the designated name information from a signer object */
-typedef pdf_pkcs7_designated_name *(pdf_pkcs7_designated_name_fn)(pdf_pkcs7_signer *signer);
-
-/* Free the resources associated with previously obtained designated name information */
-typedef void (pdf_pkcs7_drop_designated_name_fn)(pdf_pkcs7_signer *signer, pdf_pkcs7_designated_name *name);
-
-/* Predict the size of the digest. The actual digest returned by create_digest will be no greater in size */
-typedef int (pdf_pkcs7_max_digest_size_fn)(pdf_pkcs7_signer *signer);
-
-/* Create a signature based on ranges of bytes drawn from a stream */
-typedef int (pdf_pkcs7_create_digest_fn)(pdf_pkcs7_signer *signer, fz_stream *in, unsigned char *digest, int *digest_len);
-
-struct pdf_pkcs7_signer_s
-{
-	pdf_pkcs7_keep_fn *keep;
-	pdf_pkcs7_drop_fn *drop;
-	pdf_pkcs7_designated_name_fn *designated_name;
-	pdf_pkcs7_drop_designated_name_fn *drop_designated_name;
-	pdf_pkcs7_max_digest_size_fn *max_digest_size;
-	pdf_pkcs7_create_digest_fn *create_digest;
-};
+typedef struct pdf_pkcs7_signer pdf_pkcs7_signer;
 
 /* Unsaved signature fields */
-typedef struct pdf_unsaved_sig_s pdf_unsaved_sig;
-
-struct pdf_unsaved_sig_s
+typedef struct pdf_unsaved_sig
 {
 	pdf_obj *field;
-	int byte_range_start;
-	int byte_range_end;
-	int contents_start;
-	int contents_end;
+	size_t byte_range_start;
+	size_t byte_range_end;
+	size_t contents_start;
+	size_t contents_end;
 	pdf_pkcs7_signer *signer;
-	pdf_unsaved_sig *next;
-};
+	struct pdf_unsaved_sig *next;
+} pdf_unsaved_sig;
 
-typedef struct pdf_rev_page_map_s pdf_rev_page_map;
-struct pdf_rev_page_map_s
+typedef struct
 {
 	int page;
 	int object;
-};
+} pdf_rev_page_map;
 
-struct pdf_document_s
+typedef struct
+{
+	int number; /* Page object number */
+	int64_t offset; /* Offset of page object */
+	int64_t index; /* Index into shared hint_shared_ref */
+} pdf_hint_page;
+
+typedef struct
+{
+	int number; /* Object number of first object */
+	int64_t offset; /* Offset of first object */
+} pdf_hint_shared;
+
+typedef struct {
+	char *key;
+	fz_xml_doc *value;
+} pdf_xfa_entry;
+
+typedef struct {
+	int count;
+	pdf_xfa_entry *entries;
+} pdf_xfa;
+
+struct pdf_document
 {
 	fz_document super;
 
@@ -604,8 +303,6 @@ struct pdf_document_s
 	int64_t file_size;
 	pdf_crypt *crypt;
 	pdf_ocg_descriptor *ocg;
-	pdf_portfolio *portfolio;
-	pdf_hotspot hotspot;
 	fz_colorspace *oi;
 
 	int max_xref_len;
@@ -617,9 +314,10 @@ struct pdf_document_s
 	pdf_xref *xref_sections;
 	pdf_xref *saved_xref_sections;
 	int *xref_index;
-	int freeze_updates;
+	int save_in_progress;
 	int has_xref_streams;
 	int has_old_style_xrefs;
+	int has_linearization_object;
 
 	int rev_page_count;
 	pdf_rev_page_map *rev_page_map;
@@ -657,18 +355,9 @@ struct pdf_document_s
 	 * These are guaranteed to lie within the region starting at
 	 * hint_shared[r]->offset of length hint_shared[r]->length
 	 */
-	struct
-	{
-		int number; /* Page object number */
-		int64_t offset; /* Offset of page object */
-		int64_t index; /* Index into shared hint_shared_ref */
-	} *hint_page;
+	pdf_hint_page *hint_page;
 	int *hint_shared_ref;
-	struct
-	{
-		int number; /* Object number of first object */
-		int64_t offset; /* Offset of first object */
-	} *hint_shared;
+	pdf_hint_shared *hint_shared;
 	int hint_obj_offsets_max;
 	int64_t *hint_obj_offsets;
 
@@ -676,13 +365,11 @@ struct pdf_document_s
 
 	pdf_lexbuf_large lexbuf;
 
-	pdf_annot *focus;
-	pdf_obj *focus_obj;
-
 	pdf_js *js;
 
-	int recalculating;
+	int recalculate;
 	int dirty;
+	int redacted;
 
 	pdf_doc_event_cb *event_cb;
 	void *event_cb_data;
@@ -692,31 +379,22 @@ struct pdf_document_s
 	fz_font **type3_fonts;
 
 	struct {
-		fz_hash_table *images;
 		fz_hash_table *fonts;
 	} resources;
 
 	int orphans_max;
 	int orphans_count;
 	pdf_obj **orphans;
+
+	pdf_xfa xfa;
 };
 
-/*
-	PDF creation
-*/
-
-/*
-	pdf_create_document: Create a blank PDF document
-*/
 pdf_document *pdf_create_document(fz_context *ctx);
 
-/*
-	Deep copy objects between documents.
-*/
-typedef struct pdf_graft_map_s pdf_graft_map;
+typedef struct pdf_graft_map pdf_graft_map;
 
 /*
-	pdf_graft_object: Return a deep copied object equivalent to the
+	Return a deep copied object equivalent to the
 	supplied object, suitable for use within the given document.
 
 	dst: The document in which the returned object is to be used.
@@ -729,7 +407,7 @@ typedef struct pdf_graft_map_s pdf_graft_map;
 pdf_obj *pdf_graft_object(fz_context *ctx, pdf_document *dst, pdf_obj *obj);
 
 /*
-	pdf_new_graft_map: Prepare a graft map object to allow objects
+	Prepare a graft map object to allow objects
 	to be deep copied from one document to the given one, avoiding
 	problems with duplicated child objects.
 
@@ -739,33 +417,46 @@ pdf_obj *pdf_graft_object(fz_context *ctx, pdf_document *dst, pdf_obj *obj);
 */
 pdf_graft_map *pdf_new_graft_map(fz_context *ctx, pdf_document *dst);
 
-/*
-	pdf_keep_graft_map: Keep a reference to a graft map object.
-*/
 pdf_graft_map *pdf_keep_graft_map(fz_context *ctx, pdf_graft_map *map);
-
-/*
-	pdf_drop_graft_map: Drop a graft map.
-*/
 void pdf_drop_graft_map(fz_context *ctx, pdf_graft_map *map);
 
 /*
-	pdf_graft_mapped_object: Return a deep copied object equivalent
+	Return a deep copied object equivalent
 	to the supplied object, suitable for use within the target
 	document of the map.
 
 	map: A map targeted at the document in which the returned
 	object is to be used.
 
-	obj: The object deep copy.
+	obj: The object to be copied.
 
 	Note: Copying multiple objects via the same graft map ensures
-	that any shared child are not duplicated more than once.
+	that any shared children are not copied more than once.
 */
 pdf_obj *pdf_graft_mapped_object(fz_context *ctx, pdf_graft_map *map, pdf_obj *obj);
 
 /*
-	pdf_page_write: Create a device that will record the
+	Graft a page (and its resources) from the src document to the
+	destination document of the graft. This involves a deep copy
+	of the objects in question.
+
+	map: A map targetted at the document into which the page should
+	be inserted.
+
+	page_to: The position within the destination document at which
+	the page should be inserted (pages numbered from 0, with -1
+	meaning "at the end").
+
+	src: The document from which the page should be copied.
+
+	page_from: The page number which should be copied from the src
+	document (pages numbered from 0, with -1 meaning "at the end").
+*/
+void pdf_graft_page(fz_context *ctx, pdf_document *dst, int page_to, pdf_document *src, int page_from);
+void pdf_graft_mapped_page(fz_context *ctx, pdf_graft_map *map, int page_to, pdf_document *src, int page_from);
+
+/*
+	Create a device that will record the
 	graphical operations given to it into a sequence of
 	pdf operations, together with a set of resources. This
 	sequence/set pair can then be used as the basis for
@@ -784,7 +475,7 @@ pdf_obj *pdf_graft_mapped_object(fz_context *ctx, pdf_graft_map *map, pdf_obj *o
 fz_device *pdf_page_write(fz_context *ctx, pdf_document *doc, fz_rect mediabox, pdf_obj **presources, fz_buffer **pcontents);
 
 /*
-	pdf_add_page: Create a pdf_obj within a document that
+	Create a pdf_obj within a document that
 	represents a page, from a previously created resources
 	dictionary and page content stream. This should then be
 	inserted into the document using pdf_insert_page.
@@ -810,7 +501,7 @@ fz_device *pdf_page_write(fz_context *ctx, pdf_document *doc, fz_rect mediabox,
 pdf_obj *pdf_add_page(fz_context *ctx, pdf_document *doc, fz_rect mediabox, int rotate, pdf_obj *resources, fz_buffer *contents);
 
 /*
-	pdf_insert_page: Insert a page previously created by
+	Insert a page previously created by
 	pdf_add_page into the pages tree of the document.
 
 	doc: The document to insert into.
@@ -824,7 +515,7 @@ pdf_obj *pdf_add_page(fz_context *ctx, pdf_document *doc, fz_rect mediabox, int
 void pdf_insert_page(fz_context *ctx, pdf_document *doc, int at, pdf_obj *page);
 
 /*
-	pdf_delete_page: Delete a page from the page tree of
+	Delete a page from the page tree of
 	a document. This does not remove the page contents
 	or resources from the file.
 
@@ -835,7 +526,7 @@ void pdf_insert_page(fz_context *ctx, pdf_document *doc, int at, pdf_obj *page);
 void pdf_delete_page(fz_context *ctx, pdf_document *doc, int number);
 
 /*
-	pdf_delete_page_range: Delete a range of pages from the
+	Delete a range of pages from the
 	page tree of a document. This does not remove the page
 	contents or resources from the file.
 
@@ -848,25 +539,15 @@ void pdf_delete_page(fz_context *ctx, pdf_document *doc, int number);
 */
 void pdf_delete_page_range(fz_context *ctx, pdf_document *doc, int start, int end);
 
-/*
-	pdf_finish_edit: Called after any editing operations
-	on a document have completed, this will tidy up
-	the document. For now this is restricted to
-	rebalancing the page tree, but may be extended
-	in the future.
-*/
-void pdf_finish_edit(fz_context *ctx, pdf_document *doc);
-
-int pdf_recognize(fz_context *doc, const char *magic);
-
-typedef struct pdf_write_options_s pdf_write_options;
+fz_text_language pdf_document_language(fz_context *ctx, pdf_document *doc);
+void pdf_set_document_language(fz_context *ctx, pdf_document *doc, fz_text_language lang);
 
 /*
 	In calls to fz_save_document, the following options structure can be used
 	to control aspects of the writing process. This structure may grow
 	in the future, and should be zero-filled to allow forwards compatibility.
 */
-struct pdf_write_options_s
+typedef struct
 {
 	int do_incremental; /* Write just the changed objects. */
 	int do_pretty; /* Pretty-print dictionaries and arrays. */
@@ -879,9 +560,14 @@ struct pdf_write_options_s
 	int do_linear; /* Write linearised. */
 	int do_clean; /* Clean content streams. */
 	int do_sanitize; /* Sanitize content streams. */
-	int continue_on_error; /* If set, errors are (optionally) counted and writing continues. */
-	int *errors; /* Pointer to a place to store a count of errors */
-};
+	int do_appearance; /* (Re)create appearance streams. */
+	int do_encrypt; /* Encryption method to use: keep, none, rc4-40, etc. */
+	int permissions; /* Document encryption permissions. */
+	char opwd_utf8[128]; /* Owner password. */
+	char upwd_utf8[128]; /* User password. */
+} pdf_write_options;
+
+extern const pdf_write_options pdf_default_write_options;
 
 /*
 	Parse option string into a pdf_write_options struct.
@@ -897,24 +583,27 @@ struct pdf_write_options_s
 pdf_write_options *pdf_parse_write_options(fz_context *ctx, pdf_write_options *opts, const char *args);
 
 /*
-	pdf_has_unsaved_sigs: Returns true if there are digital signatures waiting to
+	Returns true if there are digital signatures waiting to
 	to updated on save.
 */
 int pdf_has_unsaved_sigs(fz_context *ctx, pdf_document *doc);
 
 /*
-	pdf_write_document: Write out the document to an output stream with all changes finalised.
+	Write out the document to an output stream with all changes finalised.
 */
 void pdf_write_document(fz_context *ctx, pdf_document *doc, fz_output *out, pdf_write_options *opts);
 
 /*
-	pdf_save_document: Write out the document to a file with all changes finalised.
+	Write out the document to a file with all changes finalised.
 */
 void pdf_save_document(fz_context *ctx, pdf_document *doc, const char *filename, pdf_write_options *opts);
 
+char *pdf_format_write_options(fz_context *ctx, char *buffer, size_t buffer_len, const pdf_write_options *opts);
+
 /*
-	pdf_can_be_saved_incrementally: Return true if the document can be saved
-	incrementally. (e.g. it has not been repaired, and it is not encrypted)
+	Return true if the document can be saved incrementally. Applying
+	redactions or having a repaired document make incremental saving
+	impossible.
 */
 int pdf_can_be_saved_incrementally(fz_context *ctx, pdf_document *doc);
 
